"""
[General documentation about the :mod:`exceptions` submodule]
"""
from __future__ import absolute_import

from . import compatibility
from .compatibility import *

__all__ = compatibility.__all__
